use std::sync::Arc;

use anyhow::format_err;
use ecomobicoin_jsonrpc::types::{self};
use ethereum_types::H256;
use fastrlp::Decodable;
use itertools::Itertools;
use jsonrpsee::{
    core::{async_trait, RpcResult},
    proc_macros::rpc,
};
use mdbx::WriteMap;
use rand::Rng;
use tracing::{debug, info};

use crate::consensus::miner::get_latest_parent_header;
use crate::{
    accessors::chain,
    consensus::{
        engine_factory,
        miner::{AuthoringParams, Miner, MinerOptions, MiningStrategy},
    },
    execution::{analysis_cache::AnalysisCache, tracer::NoopTracer},
    kv::MdbxWithDirHandle,
    models::{
        behavior::BehaviorWithSignature, BlockNumber, ChainSpec, ExecutedBlock,
        MessageWithSignature,
    },
};
use crate::{kv::tables, p2p::node::Node};

use super::helpers;

#[rpc(server, client, namespace = "emc")]
pub trait MinerApi {
    #[method(name = "mine")]
    async fn mine(&self, behavior: BehaviorWithSignature) -> RpcResult<ExecutedBlock>;
    #[method(name = "send_raw_behavior")]
    async fn send_raw_behavior(&self, behavior_data: types::Bytes) -> RpcResult<H256>;
    #[method(name = "send_raw_transaction")]
    async fn send_raw_transaction(&self, transaction_data: types::Bytes) -> RpcResult<H256>;

    // SendTransaction signs and submits a transaction (supposing the private key is managed by the node so not used yet by Ecomobicoin)
    // #[method(name = "send_behavior")]
    // async fn send_behavior(&self, call_data: types::BehaviorCall) -> RpcResult<bool>;
    // #[method(name = "send_transaction")]
    // async fn send_transaction(&self, call_data: types::MessageCall) -> RpcResult<bool>;
}

pub struct MinerRpc {
    pub db: Arc<MdbxWithDirHandle<WriteMap>>,
    pub chain_spec: ChainSpec,
    pub node: Arc<Node>,
}

#[async_trait]
impl MinerApiServer for MinerRpc {
    async fn mine(&self, behavior: BehaviorWithSignature) -> RpcResult<ExecutedBlock> {
        let db1 = Arc::clone(&self.db);
        let db2 = Arc::clone(&self.db);
        let db3 = Arc::clone(&self.db);

        let tx = db3.begin()?;
        let last_header = get_latest_parent_header(&tx);
        let last_block_number = {
            if last_header.is_err() {
                BlockNumber(0)
            } else {
                let header = last_header.unwrap();
                header.number
            }
        };
        let chain_spec = self.chain_spec.clone();
        let block_spec = self.chain_spec.collect_block_spec(last_block_number);
        let node = self.node.clone();

        tokio::task::spawn_blocking(move || {
            let mut analysis_cache = AnalysisCache::default();
            let mut tracer = NoopTracer;

            // engine already created in `akula.rs`, should we re-use it ?
            let engine = engine_factory(Some(db1), chain_spec, None)?;

            let miner_addr = behavior.recover_sender()?;

            debug!(
                "mine() with Behavior:{:?} from {:?}",
                behavior.clone(),
                miner_addr.clone()
            );

            let mut miner = Miner::<WriteMap>::new(
                AuthoringParams {
                    author: miner_addr,
                    extra_data: b"LIMOS_CNRS_UCA".to_vec().into(),
                },
                MinerOptions {
                    dry_run: false,
                    strategy: MiningStrategy::EqualProportions,
                },
                &mut tracer,
                &mut analysis_cache,
                engine,
                &block_spec,
                db2,
                node,
            );
            //MINE
            let executed_block: ExecutedBlock = miner.start_from_pob_m(behavior)?;

            debug!(
                "Mined block #{:?}, hash {:?}",
                executed_block.header.number,
                executed_block.header.hash()
            );
            debug!("Content of mined block {:?}", executed_block);
            Ok(executed_block)
        })
        .await
        .unwrap_or_else(helpers::joinerror_to_result)
    }

    async fn send_raw_behavior(&self, behavior_data: types::Bytes) -> RpcResult<H256> {
        let db2 = Arc::clone(&self.db);

        tokio::task::spawn_blocking(move || {
            let signed_behavior = BehaviorWithSignature::decode(&mut &*behavior_data.0)
                .map_err(|e| format_err!("failed to decode bx: {e}"))?;

            // let sender = signed_behavior.recover_sender()?;

            let txn = db2.begin_mutable().unwrap();

            let bx_id: u64 = rand::thread_rng().gen();
            let bx_hash = signed_behavior.hash();
            chain::bx_pending::write(&txn, bx_id, &[signed_behavior]).unwrap();
            info!("Wrote behavior {:?}, {:?} to pool", bx_id, bx_hash);
            txn.commit()?;
            Ok(bx_hash)
        })
        .await
        .unwrap_or_else(helpers::joinerror_to_result)
    }

    async fn send_raw_transaction(&self, transaction_data: types::Bytes) -> RpcResult<H256> {
        let db2: Arc<MdbxWithDirHandle<WriteMap>> = Arc::clone(&self.db);

        tokio::task::spawn_blocking(move || {
            let signed_message =
                <MessageWithSignature as fastrlp::Decodable>::decode(&mut &*transaction_data.0)
                    .map_err(|e| format_err!("failed to decode tx: {e}"))?;
            // let sender = signed_message.recover_sender()?;

            let txn = db2.begin_mutable().unwrap();
            let db_entries = txn
                .cursor(tables::PendingTransaction)?
                .walk(None)
                .collect_vec();
            let mut db_state: Vec<MessageWithSignature> = vec![];
            for entry in db_entries {
                db_state.push(entry.unwrap().1);
            }
            if !db_state.contains(&signed_message) {
                let tx_id: u64 = rand::thread_rng().gen();
                let tx_hash = signed_message.hash();
                chain::tx_pending::write(&txn, tx_id, &[signed_message]).unwrap();
                info!(
                    "Added transaction to pool id:{:?}, hash: {:?}",
                    tx_id, tx_hash
                );
                txn.commit()?;
                Ok(tx_hash)
            } else {
                info!("Transaction already present in pool");
                txn.commit().expect("Failed to save, no duplicate policy");
                Ok(H256::zero())
            }
        })
        .await
        .unwrap_or_else(helpers::joinerror_to_result)
    }
}
